import sys
import os
from datetime import datetime

import numpy as np
import pandas as pd
from statsmodels.stats.multitest import multipletests


### Constants ###

ALL_NTS_LIST = list('ACGT')
ALL_NTS_SET = set(ALL_NTS_LIST)


### Project Functions ###

def log(message, end = '\n'):
    print('FABRIC|PID-%s [%s]: %s' % (os.getpid(), datetime.now(), message), end = end)
    sys.stdout.flush()
    
    
### Statistics ###
    
def multipletests_with_nulls(values, method = 'fdr_bh'):
    
    significance = np.full(len(values), False, dtype = bool)
    qvals = np.full(len(values), np.nan, dtype = float)
    mask = pd.notnull(values)
    
    if mask.any():
        significance[np.array(mask)], qvals[np.array(mask)], _, _ = multipletests(values[mask], method = method)
    
    return significance, qvals



from itertools import product

import numpy as np
import pandas as pd
from scipy.stats import ttest_ind
from statsmodels.stats.api import DescrStatsW, CompareMeans

from .util import ALL_NTS_LIST, log, multipletests_with_nulls
from .gene_score_dist import get_gene_score_distribution
from .gene_testing import test_gene

def analyze_variant_effects(variant_effects, genes, gene_score_models, other_variant_effects_to_compare_against = None):

    varaint_effects_by_gene = variant_effects.groupby('gene_index')
    
    if other_variant_effects_to_compare_against is not None:
        other_variant_effects_by_gene = other_variant_effects_to_compare_against.groupby('gene_index')
        
    results = []

    for i, (gene_index, gene_variant_effects) in enumerate(varaint_effects_by_gene):
        
        log('Analyzing gene %d/%d...' % (i, len(varaint_effects_by_gene)), end = '\r')
        gene = genes.loc[gene_index]
        gene_score_model = gene_score_models[gene_index]
        overall_bg_score_dist = get_gene_score_distribution(gene_score_model, get_nt_substitution_counts(gene_variant_effects))
        missense_bg_score_dist = get_gene_score_distribution(gene_score_model, \
                get_nt_substitution_counts(gene_variant_effects[gene_variant_effects['effect_type'] == 'missense']))
        gene_results = pd.concat([gene, test_gene(gene['cds_len'], gene_variant_effects, overall_bg_score_dist, missense_bg_score_dist)]).rename(gene_index)
        
        if other_variant_effects_to_compare_against is not None:
            if gene_index in other_variant_effects_by_gene:
                
                gene_other_variant_effects = other_variant_effects_by_gene.get_group(gene_index)
                other_bg_score_dist = get_gene_score_distribution(gene_score_model, get_nt_substitution_counts(gene_other_variant_effects))
                        
                reference_z_scores = (gene_variant_effects['effect_scores'] - overall_bg_score_dist.mean()) / overall_bg_score_dist.std()
                other_z_scores = (gene_other_variant_effects['effect_score'] - other_bg_score_dist.mean()) / other_bg_score_dist.std()
                        
                gene_results['diff_z_scores_ci'] = '%f-%f' % CompareMeans(DescrStatsW(reference_z_scores), DescrStatsW(other_z_scores)).tconfint_diff()
                _, gene_results['diff_pval'] = ttest_ind(reference_z_scores, other_z_scores)
            else:
                gene_results['diff_z_scores_ci'] = np.nan
                gene_results['diff_pval'] = np.nan
            
        results.append(gene_results)
        
    results = pd.DataFrame(results).sort_index()
    results['overall_fdr_significance'], results['overall_fdr_qval'] = multipletests_with_nulls(results['overall_pval'])
    results['missense_fdr_significance'], results['missense_fdr_qval'] = multipletests_with_nulls(results['missense_pval'])
    results['types_chi2_fdr_significance'], results['types_chi2_fdr_qval'] = multipletests_with_nulls(results['types_chi2_pval'])
    return results
    
def get_nt_substitution_counts(variants):
    return variants.groupby(['ref', 'alt']).size().reindex(list(product(ALL_NTS_LIST, ALL_NTS_LIST))).fillna(0).astype(int).unstack()



import re
from collections import Counter

import numpy as np
import pandas as pd
        
def extract_variants(vcf_file_handler, only_pass = False, min_AF = 0):
    
    n_filtered_variants = Counter()
    variants = []
    
    for line in vcf_file_handler:
        if line.startswith('##'):
            continue
        elif line.startswith('#'):
            headers = line[1:].strip().split('\t')
        else:
            
            AFs = np.array(list(map(float, _AF_PATTERN.search(line).group(1).split(','))))
            AF_mask = (AFs >= min_AF)
            
            if AF_mask.any():
                
                variant_data = dict(zip(headers, line.strip().split('\t')))
                
                if not only_pass or variant_data['FILTER'] == 'PASS':
                    
                    alts = np.array(variant_data['ALT'].split(','))
                    
                    for alt, AF in zip(alts[AF_mask], AFs[AF_mask]):
                        variants.append([variant_data['ID'], variant_data['CHROM'], int(variant_data['POS']), \
                                variant_data['REF'], alt, AF, float(variant_data['QUAL']), variant_data['FILTER']])
                else:
                    n_filtered_variants['FILTER is not "PASS"'] += 1
            else:
                n_filtered_variants['AF below %f' % min_AF] += 1
    
    variants = pd.DataFrame(variants, columns = ['id', 'chrom', 'pos', 'ref', 'alt', 'AF', 'qual', 'filter'])
    assert (variants['ref'] != variants['alt']).all()
    return variants, n_filtered_variants

_AF_PATTERN = re.compile('AF=(.*?);')



import os

def get_parser_file_type(parser, must_exist = False):

    def _file_type(path):
    
        path = os.path.expanduser(path)
    
        if must_exist:
            if not os.path.exists(path):
                parser.error('File doesn\'t exist: %s' % path)
            elif not os.path.isfile(path):
                parser.error('Not a file: %s' % path)
            else:
                return path
        else:
        
            dir_path = os.path.dirname(path)
        
            if not os.path.exists(dir_path):
                parser.error('Parent directory doesn\'t exist: %s' % dir_path)
            else:
                return path
    
    return _file_type

def get_parser_directory_type(parser, create_if_not_exists = False):
    
    def _directory_type(path):
    
        path = os.path.expanduser(path)
    
        if not os.path.exists(path):
            if create_if_not_exists:
                if os.path.exists(os.path.dirname(path)):
                    os.mkdir(path)
                    return path
                else:
                    parser.error('Cannot create empty directory (parent directory doesn\'t exist): %s' % path)
            else:
                parser.error('Path doesn\'t exist: %s' % path)
        elif not os.path.isdir(path):
            parser.error('Not a directory: %s' % path)
        else:
            return path
        
    return _directory_type


import numpy as np
from scipy.stats import rv_discrete

def get_gene_score_distribution(gene_score_model, nt_substitution_counts):
    nt_substitution_freqs = nt_substitution_counts / nt_substitution_counts.sum().sum()
    return GeneScoreDistribution(gene_score_model, nt_substitution_freqs)
    
class GeneScoreDistribution(object):

    def __init__(self, gene_score_model, substitution_freqs_matrix):
        self.gene_score_model = gene_score_model
        self.substitution_freqs_matrix = substitution_freqs_matrix
        self._decompose_substitution_freqs()
        
    def is_empty(self):
        return self._empty
        
    def get_type_freqs(self):
        if self._empty:
            return np.array(3 * [np.nan])
        else:
            type_freqs_per_substitution = np.array([substitution_score_dist.get_type_freqs() for substitution_score_dist in self._substitution_score_dists])
            return np.dot(self._freqs.reshape(1, -1), type_freqs_per_substitution).flatten()
        
    def sample(self, n, only_missense = False):
        if self._empty:
            return np.array(n * [np.nan])
        else:
            substitution_counts = np.random.multinomial(n, self._freqs)
            samples = np.concatenate([substitution_score_dist.sample(count, only_missense = only_missense) \
                    for count, substitution_score_dist in zip(substitution_counts, self._substitution_score_dists)])
            np.random.shuffle(samples)
            return samples
            
    def mean(self, only_missense = False):
        if self._empty:
            return np.nan
        else:
            return (self._freqs * self._get_substitution_means(only_missense = only_missense)).sum()
        
    def var(self, only_missense = False):
        if self._empty:
            return np.nan
        else:
            
            mean_of_substitution_vars = (self._freqs * self._get_substitution_vars(only_missense = only_missense)).sum()
            substitution_means_dist = rv_discrete(values = (self._get_substitution_means(only_missense = only_missense), self._freqs))
            var_of_substitution_means = substitution_means_dist.var()
            
            # According the the law of total variance.
            return mean_of_substitution_vars + var_of_substitution_means
        
    def std(self, only_missense = False):
        return np.sqrt(self.var(only_missense = only_missense))
        
    def to_bins(self, n_bins, only_missense = False, reversed = False):
        substitution_bins = [substitution_score_dist.to_bins(n_bins, only_missense = only_missense, reversed = reversed) for \
                substitution_score_dist in self._substitution_score_dists]
        return np.dot(self._freqs, substitution_bins)
        
    def _get_substitution_means(self, only_missense):
        return np.array([substitution_score_dist.mean(only_missense = only_missense) for substitution_score_dist in self._substitution_score_dists])
        
    def _get_substitution_vars(self, only_missense):
        return np.array([substitution_score_dist.var(only_missense = only_missense) for substitution_score_dist in self._substitution_score_dists])
        
    def _decompose_substitution_freqs(self):
        
        substitution_freqs = self.substitution_freqs_matrix.stack()
        substitution_freqs = substitution_freqs[substitution_freqs != 0]
        
        if len(substitution_freqs) == 0:
            self._empty = True
        else:
            self._empty = False
            self._substitutions, self._freqs = zip(*substitution_freqs.to_dict().items())
            self._freqs = np.array(self._freqs)
            self._substitution_score_dists = [self.gene_score_model.substitution_score_distributions[substitution] \
                    for substitution in self._substitutions]
    
class GeneScoreModel(object):

    def __init__(self, variants_df):
        self.substitution_score_distributions = {}
        self._parse_data(variants_df)
    
    def _parse_data(self, variants_df):
        for (ref, alt), substitution_variants in variants_df.groupby(['ref', 'alt']):
            n_total = len(substitution_variants)
            n_synonymous = (substitution_variants['effect_type'] == 'synonymous').sum()
            n_nonsense = (substitution_variants['effect_type'] == 'nonsense').sum()
            n_missense = (substitution_variants['effect_type'] == 'missense').sum()
            missense_scores = substitution_variants.loc[substitution_variants['effect_type'] == 'missense', 'effect_score'].values
            self.substitution_score_distributions[(ref, alt)] = SubstitutionScoreDistribution(n_total, n_synonymous, n_nonsense, \
                    n_missense, missense_scores)

class SubstitutionScoreDistribution(object):

    def __init__(self, n_total, n_synonymous, n_nonsense, n_missense, missense_scores):
        assert n_total == n_synonymous + n_nonsense + n_missense
        assert len(missense_scores) == n_missense
        self.f_synonymous = n_synonymous / n_total
        self.f_nonsense = n_nonsense / n_total
        self.f_missense = n_missense / n_total
        self.missense_scores = np.array(missense_scores)
        
    def get_type_freqs(self):
        return np.array([self.f_synonymous, self.f_missense, self.f_nonsense])
        
    def sample(self, n, only_missense = False):
        if only_missense:
            return self._sample_missense(n)
        else:

            rand = np.random.uniform(0, 1, n)
            synonymous_mask = (rand <= self.f_synonymous)
            missense_mask = ((rand > self.f_synonymous) & (rand <= self.f_synonymous + self.f_missense))
            
            samples = np.zeros(n)
            samples[synonymous_mask] = 1
            samples[missense_mask] = self._sample_missense(missense_mask.sum())
            
            return samples
            
    def mean(self, only_missense = False):
        if only_missense:
            return self.missense_scores.mean()
        else:
            return self.f_synonymous + self.f_missense * self.missense_scores.mean()
        
    def var(self, only_missense = False):
        if only_missense:
            return self.missense_scores.var()
        else:
            
            mean_of_type_vars = self.f_missense * self.missense_scores.var()
            type_means_dist = rv_discrete(values = ([1, self.missense_scores.mean(), 0], self.get_type_freqs()))
            var_of_type_means = type_means_dist.var()
            
            # According the the law of total variance.
            return mean_of_type_vars + var_of_type_means
        
    def std(self, only_missense = False):
        return np.sqrt(self.var(only_missense = only_missense))
        
    def to_bins(self, n_bins, only_missense = False, reversed = False):
        
        bins = self._missense_to_bins(n_bins, reversed = reversed)
        
        if not only_missense:
            
            bins *= self.f_missense
            
            if reversed:
                bins[0] += self.f_synonymous
                bins[-1] += self.f_nonsense
            else:
                bins[0] += self.f_nonsense
                bins[-1] += self.f_synonymous
            
        assert np.isclose(bins.sum(), 1)
        return bins
        
    def _sample_missense(self, n):
        return np.random.choice(self.missense_scores, size = n, replace = True)
        
    def _missense_to_bins(self, n_bins, reversed):
        
        scores = 1 - self.missense_scores if reversed else self.missense_scores
        prob_per_score = 1 / len(scores)
        bin_indices = (n_bins * scores).astype(int)
        bins = np.zeros(n_bins + 1)
        
        for bin_index in bin_indices:
            bins[bin_index] += prob_per_score
            
        return bins



import numpy as np
import pandas as pd
from scipy.stats import chisquare

def test_gene(cds_len, observed_effects, overall_bg_score_dist, missense_bg_score_dist):
    
    n_obs = len(observed_effects)
    mutations_per_nt = n_obs / cds_len
    meta_results = pd.Series([n_obs, mutations_per_nt], index = ['n_obs', 'mutations_per_nt'])
        
    overall_score_results = test_gene_scores(observed_effects, overall_bg_score_dist, False)
    missense_score_results = test_gene_scores(observed_effects, missense_bg_score_dist, True)
    
    observed_type_counts = observed_effects['effect_type'].value_counts().reindex(['synonymous', 'missense', 'nonsense']).fillna(0).astype(int).values
    expected_type_freqs = overall_bg_score_dist.get_type_freqs()
    
    if (expected_type_freqs == 0).any():
        types_chi2_pval = np.nan
    else:
        _, types_chi2_pval = chisquare(observed_type_counts, observed_type_counts.sum() * expected_type_freqs)
    
    type_results = pd.Series([types_chi2_pval, list(observed_type_counts), list(expected_type_freqs)], \
            index = ['types_chi2_pval', 'observed_type_counts', 'expected_type_freqs'])
    
    return pd.concat([meta_results, overall_score_results, missense_score_results, type_results])
            
def test_gene_scores(observed_effects, bg_score_dist, only_missense):
    
    if only_missense:
        observed_scores = observed_effects.loc[observed_effects['effect_type'] == 'missense', 'effect_score']
    else:
        observed_scores = observed_effects['effect_score']
    
    bg_mean = bg_score_dist.mean(only_missense = only_missense)
    bg_std = bg_score_dist.std(only_missense = only_missense)
    
    observed_avg = observed_scores.mean()
    mean_deviation = observed_avg - bg_mean
    z_value = mean_deviation / bg_std
    pval = calc_two_tailed_binning_pval(observed_scores, bg_score_dist, only_missense)
        
    results = pd.Series([bg_mean, bg_std, observed_avg, mean_deviation, z_value, pval], \
            index = ['bg_mean', 'bg_std', 'observed_avg', 'mean_deviation', 'z_value', 'pval'])
            
    if only_missense:
        results.rename(lambda column: 'missense_%s' % column, inplace = True)
    else:
        results.rename(lambda column: 'overall_%s' % column, inplace = True)
            
    return results

def calc_two_tailed_binning_pval(obs_scores, bg_score_dist, only_missense, n_bins = 100):
    
    if bg_score_dist.is_empty():
        return np.nan
    
    if obs_scores.mean() <= bg_score_dist.mean():
        one_tail_func = calc_left_tail_binning_pval
    else:
        one_tail_func = calc_right_tail_binning_pval
        
    return min(1, 2 * one_tail_func(obs_scores, bg_score_dist, only_missense, n_bins = n_bins))
    
def calc_left_tail_binning_pval(obs_scores, bg_score_dist, only_missense, n_bins = 100):
    obs_sum_bin = (n_bins * obs_scores).astype(int).sum()
    bg_bins = bg_score_dist.to_bins(n_bins, only_missense = only_missense)
    bg_sum_bins = iid_sum_bins(bg_bins, len(obs_scores), max_bin = obs_sum_bin)
    return bg_sum_bins[:(obs_sum_bin + 1)].sum()
    
def calc_right_tail_binning_pval(obs_scores, bg_score_dist, only_missense, n_bins = 100):
    reversed_obs_scores = 1 - obs_scores
    reversed_obs_sum_bin = (n_bins * reversed_obs_scores).astype(int).sum()
    reversed_bg_bins = bg_score_dist.to_bins(n_bins, only_missense = only_missense, reversed = True)
    reversed_bg_sum_bins = iid_sum_bins(reversed_bg_bins, len(obs_scores), max_bin = reversed_obs_sum_bin)
    return reversed_bg_sum_bins[:(reversed_obs_sum_bin + 1)].sum()
    
def iid_sum_bins(bins, n, max_bin = None):
    
    n_as_bits = list(map(int, bin(n)[2:][::-1]))
    
    if n_as_bits[0] == 1:
        sum_bins = bins
    else:
        sum_bins = None
        
    for bit in n_as_bits[1:]:
        
        bins = np.convolve(bins, bins)
        
        if max_bin is not None:
            bins = bins[:(max_bin + 1)]
        
        if bit == 1:
            if sum_bins is None:
                sum_bins = bins
            else:
                
                sum_bins = np.convolve(sum_bins, bins)
                
                if max_bin is not None:
                    sum_bins = sum_bins[:(max_bin + 1)]
                
    return sum_bins



from setuptools import setup
        
setup(
    name = 'fabric-genetics',
    version = '1.2.0',
    description = 'FABRIC (Functional Alteration Bias Recovery In Coding-regions) is a framework for detecting genes showing functional alteration bias.',
    long_description = 'See: https://github.com/nadavbra/fabric',
    url = 'https://github.com/nadavbra/fabric',
    author = 'Nadav Brandes',
    author_email  ='nadav.brandes@mail.huji.ac.il',
    license = 'MIT',
    packages = ['fabric'],
    scripts = [
        'bin/fabric',
        'bin/vcf_to_csv',
    ],
    install_requires = [
        'numpy',
        'scipy',
        'pandas',
        'statsmodels',
    ],
)
